{
   uImage, image loading/saving and manipulation
   Copyright (C) 2007. Dejan Boras

   This file is part of dImage.

   dImage is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   dImage is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with dImage.  If not, see <http://www.gnu.org/licenses/>.

   Started On:    26.05.2007.
   Last Update:   15.04.2011.
}

{$MODE OBJFPC}{$H+}{$I-}
UNIT uImage;

{Intended to provide a unified 'interface' and routines for manipulating and
loading images of various formats.}

INTERFACE

   USES uStd, {$IFNDEF NO_LOGGING}uLog, {$ENDIF}uError, uColors,
     uFileHandlers, uFile, uFileStd;

CONST
   imgcName = 'dImage';

   {determines if the loading routines should store filename into the record}
   imgcStoreFileNames: boolean = true;
   doLog: boolean = false; {whether or not to perform logging}
   doLogNameAlways: boolean = false; {always log the name of the image being loaded}

   {ERROR CONSTANTS}
   imgeGENERAL                = $0100;
   imgeLOADER_NOT_FOUND       = $0101;
   imgeWRITER_NOT_FOUND       = $0102;
   imgeFILEHANDLER            = $0103;

   imgeUNSUPPORTED                  = $0104;
   imgeUNSUPPORTED_PIXF             = $0105;
   imgeUNSUPPORTED_DEPTH            = $0106;
   imgeUNSUPPORTED_CHANNEL_ORDER    = $0107;
   imgeUNSUPPORTED_CHANNELS         = $0108;
   imgeUNSUPPORTED_BPC              = $0109;
   imgeUNSUPPORTED_BPP              = $010A;
   imgeUNSUPPORTED_COLORMAP         = $010B;
   imgeUNSUPPORTED_COMPRESSION      = $010C;

   imgeINVALID_PIXF                 = $010D;
   imgeINVALID_DEPTH                = $010E;
   imgeINVALID_CHANNEL_ORDER        = $010F;
   imgeINVALID_CHANNELS             = $0110;
   imgeINVALID_BPC                  = $0111;
   imgeINVALID_BPP                  = $0112;
   imgeINVALID_COLORMAP             = $0113;
   imgeINVALID_COMPRESSION          = $0114;
   imgeINVALID_DIMENSIONS           = $0115;

   {GENERAL CONSTANTS}
   imgcNone                   = $00; {refers to anything that supports none}

   {PIXEL FORMAT IDs}

   {INDEXED RGB/A}
   PIXF_INDEX_RGB_4     = $00; PIXF_INDEX_4 = $00;
   PIXF_INDEX_RGBA_4    = $01;

   PIXF_INDEX_RGB_8     = $02; PIXF_INDEX_8 = $02;
   PIXF_INDEX_RGBA_8    = $03;

   {INDEXED BGR/A}
   PIXF_INDEX_BGR_4     = $04;
   PIXF_INDEX_BGR_8     = $05;
   PIXF_INDEX_BGRA_4    = $06;
   PIXF_INDEX_BGRA_8    = $07;

   {PACKED RGB/A}
   PIXF_RGB_5_6_5       = $08;
   PIXF_RGBA_5_5_5_1    = $09;
   {PACKED BGR/A}
   PIXF_BGR_5_6_5       = $0A;
   PIXF_BGRA_5_5_5_1    = $0B;

   {RGB}
   PIXF_RGB             = $0C;
   PIXF_RGBA            = $0D;
   {BGR}
   PIXF_BGR             = $0E;
   PIXF_BGRA            = $0F;

   {48-bit}
   PIXF_RGB_48          = $10;
   PIXF_RGBA_48         = $11;
   PIXF_BGR_48          = $12;
   PIXF_BGRA_48         = $13;

   {INVERTED}
   PIXF_ARGB            = $14;
   PIXF_ABGR            = $15;

   {GREYSCALE}
   PIXF_GREYSCALE_8     = $16;
   PIXF_GREYSCALE_16    = $17;
   PIXF_GREYSCALE_32    = $18;

   {SPECIAL TYPES}
   PIXF_ALPHA           = $19;

   {maximum pixel format ID}
   PIXF_MAXIMUM         = $19;

   {unsupported pixel format}
   PIXF_UNSUPPORTED     = $FF;

   {COMPRESSION TYPES}
   {use imgcNone to indicate no compression}
   imgcCOMPRESSION_NONE    = $0000;
   imgcCOMPRESSION_RLE     = $0001;

   {IMAGE ORIGINS}
   {image origin bitmasks}
   imgcBT_ORIGIN_VERTICAL     = $01; {vertical origin, if 0 top, if 1 bottom}
   imgcBT_ORIGIN_HORIZONTAL   = $02; {horizontal origin, if 0 left, if 1 right}

   imgcORIGIN_BL      = imgcBT_ORIGIN_VERTICAL;
   imgcORIGIN_BR      = imgcBT_ORIGIN_VERTICAL or imgcBT_ORIGIN_HORIZONTAL;
   imgcORIGIN_TL      = $00;
   imgcORIGIN_TR      = imgcBT_ORIGIN_HORIZONTAL;
   imgcORIGIN_DEFAULT: longword = imgcORIGIN_TL;

   {color channel orders}
   PIXF_COLOR_CHANNEL_ORDER_RGB    = 0;
   PIXF_COLOR_CHANNEL_ORDER_BGR    = 1;
   PIXF_COLOR_CHANNEL_ORDER_MONO   = 2; {greyscale or monochromatic colors}
   PIXF_COLOR_CHANNEL_ORDER_DEFAULT: longint = PIXF_COLOR_CHANNEL_ORDER_RGB;

   {set the image to the default origin while loading}
   imgcSetToDefaultOrigin: boolean = false;
   {set the image to default color channel order when loading}
   imgcSetToDefaultPixelFormat: boolean = false;
   imgcDefaultPixelFormat: longint = PIXF_RGB;
   imgcDefaultPixelFormatAlpha: longint = PIXF_RGBA;
   {sets the image to the default color channel order}
   imgcSetToDefaultColorChannelOrder: boolean = false;

   {CHANNELS}
   imgccR   = $01;
   imgccG   = $02;
   imgccB   = $04;
   imgccA   = $08;
   imgcc0   = $10;
   imgcc1   = $20;
   imgcc2   = $40;
   imgcc3   = $80;

   {2 color channel constants, mostly used as a parameter for the color channel swap routine.}
   imgccRG  = imgccR or imgccG;
   imgccRB  = imgccR or imgccB;
   imgccGB  = imgccG or imgccB;
   {same as above three}
   imgccGR  = imgccRG;
   imgccBR  = imgccRB;
   imgccBG  = imgccGB;

   {5-bit to 8-bit color conversion table}
   imgc5To8Bits: array[0..31] of byte =
      (000,008,016,025,033,041,049,058,
       066,074,082,090,099,107,115,123,
       132,140,148,156,165,173,181,189,
       197,206,214,222,230,239,247,255);

   imgcFILE_TYPE_NORMAL = 00;
   imgcFILE_TYPE_TEXT   = 01;
   imgcFILE_TYPE_MEMORY = 02;

   {image properties}
   imgcPROPERTIES_PAL_EXTERNAL         = $0001; {external palette}

   imgcPIXFDepth: array[0..PIXF_MAXIMUM] of longint =
   {INDEX_RGB}
   (4,{PIXF_INDEX_RGB_4, PIXF_INDEX_4}
    4,{PIXF_INDEX_RGBA_4}
    8, {PIXF_INDEX_RGB_8, PIXF_INDEX_8}
    8, {PIXF_INDEX_RGBA_8}
    {INDEX_BGR}
    4,{PIXF_INDEX_BGR_4}
    4,{PIXF_INDEX_BGRA_4}
    8, {PIXF_INDEX_BGR_8}
    8, {PIXF_INDEX_BGRA_8}
    {PACKED RGB}
    16, {PIXF_RGB_5_6_5}
    16, {PIXF_RGB_5_5_5_1}
    16, {PIXF_BGR_5_6_5}
    16, {PIXF_BGR_5_5_5_1}
    {RGB/A}
    24, {PIXF_RGB}
    32, {PIXF_RGBA}
    {BGR/A}
    24, {PIXF_BGR}
    32, {PIXF_BGRA}
    {48 BIT}
    48, {PIXF_RGB_48}
    48, {PIXF_RGBA_48}
    48, {PIXF_BGR_48}
    48, {PIXF_BGRA_48}
    {A/RGB}
    32, {PIXF_ARGB}
    {A/BGR}
    32, {PIXF_ABGR}
    8, {PIXF_GREYSCALE_8}
    16, {PIXF_GREYSCALE_16}
    32, {PIXF_GREYSCALE_32}
    8 {PIXF_ALPHA}
    );

   imgcPIXFChannels: array[0..PIXF_MAXIMUM] of longint =
   {INDEXED RGB}
   (3, {PIXF_INDEX_RGB_4, PIXF_INDEX_4}
    4, {PIXF_INDEX_RGBA_4}
    3, {PIXF_INDEX_RGB_8, PIXF_INDEX_8}
    4, {PIXF_INDEX_RGBA_8}
    {INDEXED BGR}
    3, {PIXF_INDEX_BGR_4}
    4, {PIXF_INDEX_BGRA_4}
    3, {PIXF_INDEX_BGR_8}
    4, {PIXF_INDEX_BGRA_8}
    {PACKED}
    3, {PIXF_RGB_5_6_5}
    4, {PIXF_RGB_5_5_5_1}
    3, {PIXF_BGR_5_6_5}
    4, {PIXF_BGR_5_5_5_1}
    {RGB/BGR}
    3, {PIXF_RGB}
    4, {PIXF_RGBA}
    3, {PIXF_BGR}
    4, {PIXF_BGRA}
    {48 Bit}
    3, {PIXF_RGB_48}
    4, {PIXF_RGBA_48}
    3, {PIXF_BGR_48}
    4, {PIXF_BGRA_48}
    {INVERTED}
    4, {PIXF_ARGB}
    4, {PIXF_ABGR}
    {GREYSCALE}
    1, {PIXF_GREYSCALE_8}
    1, {PIXF_GREYSCALE_16}
    1,  {PIXF_GREYSCALE_32}
    {SPECIAL TYPES}
    1 {PIXF_ALPHA}
    );

   imgcPIXFNames: array[0..PIXF_MAXIMUM] of string =
   {INDEX_RGB}
   ('PIXF_INDEX_RGB_4',
    'PIXF_INDEX_RGBA_4',
    'PIXF_INDEX_RGB_8',
    'PIXF_INDEX_RGBA_8',
    {INDEX_BGR}
    'PIXF_INDEX_BGR_4',
    'PIXF_INDEX_BGRA_4',
    'PIXF_INDEX_BGR_8',
    'PIXF_INDEX_BGRA_8',
    {PACKED RGB}
    'PIXF_RGB_5_6_5',
    'PIXF_RGB_5_5_5_1',
    'PIXF_BGR_5_6_5',
    'PIXF_BGR_5_5_5_1',
    {RGB/A}
    'PIXF_RGB',
    'PIXF_RGBA',
    {BGR/A}
    'PIXF_BGR',
    'PIXF_BGRA',
    {48 BIT}
    'PIXF_RGB_48',
    'PIXF_RGBA_48',
    'PIXF_BGR_48',
    'PIXF_BGRA_48',
    {A/RGB}
    'PIXF_ARGB',
    {A/BGR}
    'PIXF_ABGR',
    {GREYSCALE}
    'PIXF_GREYSCALE_8',
    'PIXF_GREYSCALE_16',
    'PIXF_GREYSCALE_32',
    {ALPHA}
    'PIXF_ALPHA'
    );

   {channel types for pixel formats}
   PIXF_CHANNELS_RGB       = 0;
   PIXF_CHANNELS_RGBA      = 1;
   PIXF_CHANNELS_BGR       = 2;
   PIXF_CHANNELS_BGRA      = 3;
   PIXF_CHANNELS_ARGB      = 4;
   PIXF_CHANNELS_ABGR      = 5;
   PIXF_CHANNELS_MONO      = 6; {this is usually greyscale}
   PIXF_CHANNELS_INVALID   = -1;

   imgcPIXFChannelTypes: array[0..PIXF_MAXIMUM] of longint =
   {INDEXED RGB}
   (PIXF_CHANNELS_RGB,  {PIXF_INDEX_RGB_4, PIXF_INDEX_4}
    PIXF_CHANNELS_RGBA, {PIXF_INDEX_RGBA_4}
    PIXF_CHANNELS_RGB,  {PIXF_INDEX_RGB_8, PIXF_INDEX_8}
    PIXF_CHANNELS_RGBA, {PIXF_INDEX_RGBA_8}
    {INDEXED BGR}
    PIXF_CHANNELS_BGR,  {PIXF_INDEX_BGR_4}
    PIXF_CHANNELS_BGRA, {PIXF_INDEX_BGRA_4}
    PIXF_CHANNELS_BGR,  {PIXF_INDEX_BGR_8}
    PIXF_CHANNELS_BGRA, {PIXF_INDEX_BGRA_8}
    {PACKED}
    PIXF_CHANNELS_RGB,  {PIXF_RGB_5_6_5}
    PIXF_CHANNELS_RGB,  {PIXF_RGB_5_5_5_1}
    PIXF_CHANNELS_BGR,  {PIXF_BGR_5_6_5}
    PIXF_CHANNELS_BGR,  {PIXF_BGR_5_5_5_1}
    {RGB/BGR}
    PIXF_CHANNELS_RGB,  {PIXF_RGB}
    PIXF_CHANNELS_RGBA, {PIXF_RGBA}
    PIXF_CHANNELS_BGR,  {PIXF_BGR}
    PIXF_CHANNELS_BGRA, {PIXF_BGRA}
    {48 Bit}
    PIXF_CHANNELS_RGB,  {PIXF_RGB_48}
    PIXF_CHANNELS_RGBA, {PIXF_RGBA_48}
    PIXF_CHANNELS_BGR,  {PIXF_BGR_48}
    PIXF_CHANNELS_BGRA, {PIXF_BGRA_48}
    {INVERTED}
    PIXF_CHANNELS_ARGB, {PIXF_ARGB}
    PIXF_CHANNELS_ABGR, {PIXF_ABGR}
    PIXF_CHANNELS_MONO, {PIXF_GREYSCALE_8}
    PIXF_CHANNELS_MONO, {PIXF_GREYSCALE_8}
    PIXF_CHANNELS_MONO, {PIXF_GREYSCALE_8}
    PIXF_CHANNELS_MONO  {PIXF_ALPHA}
    );

   imgcPIXFColorChannelOrder: array[0..PIXF_MAXIMUM] of longint =
   {INDEXED RGB}
   (PIXF_COLOR_CHANNEL_ORDER_RGB, {PIXF_INDEX_RGB_4, PIXF_INDEX_4}
    PIXF_COLOR_CHANNEL_ORDER_RGB, {PIXF_INDEX_RGBA_4}
    PIXF_COLOR_CHANNEL_ORDER_RGB, {PIXF_INDEX_RGB_8, PIXF_INDEX_8}
    PIXF_COLOR_CHANNEL_ORDER_RGB, {PIXF_INDEX_RGBA_8}
    {INDEXED BGR}
    PIXF_COLOR_CHANNEL_ORDER_BGR, {PIXF_INDEX_BGR_4}
    PIXF_COLOR_CHANNEL_ORDER_BGR, {PIXF_INDEX_BGRA_4}
    PIXF_COLOR_CHANNEL_ORDER_BGR, {PIXF_INDEX_BGR_8}
    PIXF_COLOR_CHANNEL_ORDER_BGR, {PIXF_INDEX_BGRA_8}
    {PACKED}
    PIXF_COLOR_CHANNEL_ORDER_RGB, {PIXF_RGB_5_6_5}
    PIXF_COLOR_CHANNEL_ORDER_RGB, {PIXF_RGB_5_5_5_1}
    PIXF_COLOR_CHANNEL_ORDER_BGR, {PIXF_BGR_5_6_5}
    PIXF_COLOR_CHANNEL_ORDER_BGR, {PIXF_BGR_5_5_5_1}
    {RGB/BGR}
    PIXF_COLOR_CHANNEL_ORDER_RGB, {PIXF_RGB}
    PIXF_COLOR_CHANNEL_ORDER_RGB, {PIXF_RGBA}
    PIXF_COLOR_CHANNEL_ORDER_BGR, {PIXF_BGR}
    PIXF_COLOR_CHANNEL_ORDER_BGR, {PIXF_BGRA}
    {48 Bit}
    PIXF_COLOR_CHANNEL_ORDER_RGB, {PIXF_RGB_48}
    PIXF_COLOR_CHANNEL_ORDER_RGB, {PIXF_RGBA_48}
    PIXF_COLOR_CHANNEL_ORDER_BGR, {PIXF_BGR_48}
    PIXF_COLOR_CHANNEL_ORDER_BGR, {PIXF_BGRA_48}
    {INVERTED}
    PIXF_COLOR_CHANNEL_ORDER_RGB, {PIXF_ARGB}
    PIXF_COLOR_CHANNEL_ORDER_BGR, {PIXF_ABGR}
    PIXF_COLOR_CHANNEL_ORDER_MONO, {PIXF_GREYSCALE_8}
    PIXF_COLOR_CHANNEL_ORDER_MONO, {PIXF_GREYSCALE_16}
    PIXF_COLOR_CHANNEL_ORDER_MONO, {PIXF_GREYSCALE_32}
    PIXF_COLOR_CHANNEL_ORDER_MONO  {PIXF_ALPHA}
    );

   imgcPIXFHasAlpha: array[0..PIXF_MAXIMUM] of boolean =
   {INDEXED RGB}
   (false,  {PIXF_INDEX_RGB_4, PIXF_INDEX_4}
    true, {PIXF_INDEX_RGBA_4}
    false,  {PIXF_INDEX_RGB_8, PIXF_INDEX_8}
    true, {PIXF_INDEX_RGBA_8}
    {INDEXED BGR}
    false,  {PIXF_INDEX_BGR_4}
    true, {PIXF_INDEX_BGRA_4}
    false,  {PIXF_INDEX_BGR_8}
    true, {PIXF_INDEX_BGRA_8}
    {PACKED}
    false,  {PIXF_RGB_5_6_5}
    false,  {PIXF_RGB_5_5_5_1}
    false,  {PIXF_BGR_5_6_5}
    false,  {PIXF_BGR_5_5_5_1}
    {RGB/BGR}
    false,  {PIXF_RGB}
    true, {PIXF_RGBA}
    false,  {PIXF_BGR}
    true, {PIXF_BGRA}
    {48 Bit}
    false,  {PIXF_RGB_48}
    true, {PIXF_RGBA_48}
    false,  {PIXF_BGR_48}
    true, {PIXF_BGRA_48}
    {INVERTED}
    true, {PIXF_ARGB}
    true,{PIXF_ABGR}
    false, {PIXF_GREYSCALE_8}
    false, {PIXF_GREYSCALE_16}
    false, {PIXF_GREYSCALE_32}
    true); {PIXF_ALPHA}

TYPE
   imgPImage = ^imgTImage;
   imgPLoaderData = ^imgTLoaderData;
   imgPWriterData = ^imgTWriterData;
   imgPPalette = ^imgTPalette;
   imgPPaletteHandler = ^imgTPaletteHandler;
   palPLoaderData = ^palTLoaderData;

   {IMAGE RECORD}
   imgTImage = record
      sName, FileName: string;
      Properties: longword;

      Width, Height, Size, Pixels, RowAlignBytes: longint;
      PixF, PixelDepth, Origin, Compression: longint;

      Image: pointer; {image pixel data}
      pal: imgPPalette;

      {multiple image support}
(*    nImages: longword;
      Images: array of record
         Width, Height: longword;
         Image: pointer;
      end;*)
   end;

   imgTLoaderData = record
      img: imgPImage;
      fPos, fileSize: fileint;
      fileType: longword;
      f: uFile.PFile;
   end;

   imgTWriterData = record
      img: imgPImage;
   end;

   imgTPalette = record
      fn: string;
      nColors: longword;
      PixF: longword;
      Size: longint;
      Data: pointer;
      sAuthor, sName, sDescription, sProgram: string;
   end;

   imgTPaletteHandler = record
      Make: procedure(var pal: imgTPalette);
      Dispose: procedure(var pal: imgTPalette);
   end;

   palTLoaderData = record
      pal: imgPPalette;
      fPos, fileSize: fileint;
      fn: string;
      fileType: longword;
      f: pointer;
   end;

VAR
   imgError: longint;

   imgLoadersInfo: fhTHandlerInfo; {image loaders}
   imgWritersInfo: fhTHandlerInfo; {image writers}
   palLoadersInfo: fhTHandlerInfo; {palette loaders}

{ ERROR HANDLING }
procedure imgeRaise(err: longint);
procedure imgErrorReset();

{ GENERAL IMAGE ROUTINES }
{initialize a image record}
procedure imgInit(var img: imgTImage);
{initialize a loader data record}
procedure imgInitRecord(var ld: imgTLoaderData); inline;
{make an image record on the heap}
function imgMake(): imgPImage;
{load a image from a file}
procedure imgLoad(var img: imgTImage; const filename: string);
{load a image from a file}
procedure imgWrite(var img: imgTImage; const filename: string);
{dispose of the image data}
procedure imgDispose(var img: imgTImage);
procedure imgDispose(var img: imgPImage);

{checks whether an image is valid}
function imgValid(var img: imgTImage): boolean;

{create a blank texture}
procedure imgCreateBlank(var img: imgTImage; w, h: longint; pf: longword);
function imgMakeBlank(w, h: longint; pf: longword): imgPImage;

{copy image from one to another}
procedure imgCopy(var src: imgTImage; var dst: imgTImage);
procedure imgCopy(var src: imgTImage; var dst: imgPImage);

{copies area of source image to destination image, provided they are of same PIXF}
procedure imgCopyArea(var src, dst: imgTImage; px, py, dx, dy, w, h: longint);

{ IMAGE INFORMATION }
{return the pixel format depth in bits}
function imgPIXFDepth(pf: longword): longint; inline;
{return the number of channels a pixel format has}
function imgPIXFChannels(pf: longword): longint; inline;
{returns the channel type of a PIXF}
function imgPIXFChannelType(pf: longword): longint; inline;
{tells whether a pixel format has an alpha channel}
function imgPIXFHasAlpha(pf: longword): boolean; inline;
{returns the color channel order for the specified PIXF}
function imgPIXFColorChannelOrder(pf: longword): longint; inline;
{return the name of the pixel format}
function imgPIXFName(pf: longword): string;

{ IMAGE PROCESSING }
{flip image verticaly}
procedure imgFlipV(var img: imgTImage);
{flip image horizontaly}
procedure imgFlipH(var img: imgTImage);

{transform the image from one pixel format to another}
procedure imgTransform(var img: imgTImage; TargetFormat: byte);

{adjust the gamma factor of an image}
procedure imgGamma(var img: imgTImage; factor: single);
{swaps the specified color channels}
function imgSwapColorChannels(var img: imgTImage; chans: longword): boolean;
{invert the image colors}
procedure imgInvertColors(var img: imgTImage);

{set the image to the target origin}
procedure imgSetOrigin(var img: imgTImage; target: longint);
{sets the image to the default origin, based on what it's origin is now}
procedure imgSetDefaultOrigin(var img: imgTImage);
{set default colors}
procedure imgSetDefaultPixelFormat(var img: imgTImage);
{set the image to the default color channel order}
procedure imgSetDefaultColorChannelOrder(var img: imgTImage);

{ PALETTE }

{initialize a imgTPalette record}
procedure palInitRecord(var pal: imgTPalette); inline;
{initialize a imgTPaletteHandler record}
procedure palInitHandlerRecord(var palh: imgTPaletteHandler); inline;

{load a palette}
procedure palLoad(var pal: imgTPalette; const filename: string);
{save a palette}
procedure palWrite(var pal: imgTPalette; const filename: string);

{initialize a palette data}
procedure palInitData(var pal: imgTPalette);
{make a palette}
procedure palMakeData(var pal: imgTPalette);
procedure palMake(var img: imgTImage);
procedure palMake(var img: imgTImage; pf: longword; nColors: longint);
function palMake(): imgPPalette;
{dispose of a palette}
procedure palDispose(var pal: imgTPalette);
procedure palDispose(var pal: imgPPalette);
procedure palDispose(var img: imgTImage);
{returns true if a palette is valid for an image}
function palValid(var img: imgTImage): boolean;
function palValid(var pal: imgTPalette): boolean;
{calculates and returns the size of the suitable palette for the image}
function palSize(var pal: imgTPalette): longint;
{swaps the specified color channels in the palette}
procedure palSwapColorChannels(var pal: imgTPalette; chans: longword);

{copy a palette}
procedure palCopy(var src: imgTPalette; var dst: imgTPalette);
procedure palCopy(var src: imgTPalette; var dst: imgPPalette);

{ HELPER ROUTINES }

{calculates aspects of image, needs to have width, height and PIXF}
procedure imgCalculate(var img: imgTImage);
{allocates memory for image contents, must have size precalculated}
procedure imgAllocate(var img: imgTImage);

{ LOADER HELPER ROUTINES }

{These routines are meant as a help for loaders to make performing
common operations easier. This also helps to reduce the size of code
when multiple loaders are used as it reduces the number of redundant
operations(operations each loader performs). They can also be used
when manually manipulating images.}

{used by loaders to calculate the image size and number of pixels}
procedure imglCalculate(var ld: imgTLoaderData);
{used by loaders to allocate memory for the image}
procedure imglAllocate(var ld: imgTLoaderData);
{used by loaders to read the pallete from the file}
procedure imglReadPalette(var ld: imgTLoaderData; padBytes: longint);
{process the image origin}
procedure imglProcessOrigin(var ld: imgTLoaderData);
{process the image colors}
procedure imglProcessFormat(var ld: imgTLoaderData);
{process the image channel order}
procedure imglProcessColorChannelOrder(var ld: imgTLoaderData);

{seeks a image file}
procedure imglSeek(var ld: imgTLoaderData; pos: fileint);
{reads in a image block}
function imglBlockRead(var ld: imgTLoaderData; var buf; size: fileint): fileint;
{reads in a image block}
function imglBlockWrite(var ld: imgTLoaderData; var buf; size: fileint): fileint;

{ LOGGING }

{log all image loaders and writes}
{$IFNDEF NO_LOGGING}
procedure imgLogHandlers();
{$ENDIF}

IMPLEMENTATION

{ ERROR HANDLING }
procedure imgeRaise(err: longint);
begin
   imgError := err;
end;

procedure imgErrorReset();
begin
   imgError := 0;
end;

{ IMAGE SUPPORT }

procedure imgInit(var img: imgTImage);
begin
   Zero(img, SizeOf(imgTImage));
end;

procedure imgInitRecord(var ld: imgTLoaderData); inline;
begin
   Zero(ld, SizeOf(imgTLoaderData));
end;

function imgMake(): imgPImage;
var
   img: imgPImage = nil;

begin
   new(img);
   if(img <> nil) then imgInit(img^)
   else imgeRaise(eNO_MEMORY);

   imgMake := img;
end;

procedure imgLoad(var img: imgTImage; const FileName: string);
var
   data: imgTLoaderData;
   fd: fhTFindData;
   fhdata: fhTData;
   f: TFile;
   fn: string;

   procedure logLoad();
   begin
      {$IFNDEF NO_LOGGING}
      if(not doLogNameAlways) then
         logWriteln(imgcName+' > (error loading)Image: '+FileName);
      {$ENDIF}
   end;

begin
   imgErrorReset();

   assert(@img <> nil, 'imgLoad received a nil image parameter.');

   fInit(f);
   {initialize}
   if(length(filename) <> 0) then begin
      {set up the loader data}
      imgInitRecord(data);
      data.f := @f;
      data.img := @img;

      {store the filename if specified}
      if(imgcStoreFileNames = true) then img.FileName := FileName;

      {$IFNDEF NO_LOGGING}
      if(doLogNameAlways) then logWriteln(imgcName+' > Loading: '+FileName);
      {$ENDIF}

      {call the file handler which will figure out what loader to call}
      fhFindHandler(FileName, imgLoadersInfo, fd);
      if(fhError <> 0) then begin
         imgeRaise(imgeFILEHANDLER); exit;
      end;

      {if a loader has been found}
      if(fd.handler <> nil) then begin
         {setup the filename}
         fn := FileName;
         if(fhcGuessExtension) and (fd.ext <> nil) then
            fn := FileName+fd.ext^.ext;
         fhdata.fn := fn;

         {open the file if instructed to do so}
         if(not fd.handler^.DoNotOpenFile) then begin
            fOpen(f, fn);
            if(fError <> 0) then begin imgeRaise(eIO); exit; end;

            data.fileSize := fSize(f);
            if(data.fileSize = 0) then begin
               fClose(f); imgeRaise(eEMPTY); exit;
            end;
         end;

         {call the loader}
         fhLoad(fd.handler^, @data, fhdata);

         {if there is no error, perform post-loading tasks}
         if(imgError = 0) then begin
            {set the origin}
            imglProcessOrigin(data);
            {set the colors}
            imglProcessFormat(data);
            {set the channel order}
            imglProcessColorChannelOrder(data);;
         end;

         {close the file if instructed to do so}
         if(not fd.handler^.DoNotOpenFile) then begin
            fClose(f); fErrorIgn();
         end;
      end else begin imgeRaise(imgeLOADER_NOT_FOUND); exit; end;
   end;
end;

procedure imgWrite(var img: imgTImage; const FileName: string);
var
   data: imgTLoaderData;
   fd: fhTFindData;
   fhdata: fhTData;
   f: TFile;
   fn: string;

   procedure logWrite();
   begin
      {$IFNDEF NO_LOGGING}
      if(not doLogNameAlways) then
         logWriteln(imgcName+' > (error writting)Image: '+FileName);
      {$ENDIF}
   end;

begin
   assert(@img <> nil, 'imgWrite received a nil image parameter.');

   fInit(f);

   {initialize}
   imgErrorReset();
   if(length(filename) <> 0) then begin
      {set up the loader data}
      imgInitRecord(data);
      data.f := @f;
      data.img := @img;

      {$IFNDEF NO_LOGGING}
      if(doLogNameAlways) then logWriteln(imgcName+' > Writting: '+FileName);
      {$ENDIF}

      {call the file handler which will figure out what loader to call}
      fhFindHandler(FileName, imgWritersInfo, fd);
      if(fhError <> 0) then begin
         imgeRaise(imgeFILEHANDLER); exit;
      end;

      {if a loader has been found}
      if(fd.handler <> nil) then begin
         {setup the filename}
         fn := FileName;
         if(fhcGuessExtension) and (fd.ext <> nil) then
            fn := FileName+fd.ext^.ext;
         fhdata.fn := fn;

         {open the file if instructed to do so}
         if(not fd.handler^.DoNotOpenFile) then begin
            fNew(f, fn);
            if(fError <> 0) then begin imgeRaise(eIO); exit; end;
         end;

         {call the loader}
         fhWrite(fd.handler^, @data, fhdata);

         {close the file if instructed to do so}
         if(not fd.handler^.DoNotOpenFile) then begin
            fClose(f); if(fError <> 0) then imgeRaise(eIO);
         end;
      end else begin imgeRaise(imgeWRITER_NOT_FOUND); exit; end;
   end;
end;


{Disposes of the memory allocated to image and it's fields.}
procedure imgDispose(var img: imgTImage);
begin
   img.FileName := '';
   img.sName := '';
   XFreeMem(img.Image);
   palDispose(img);

   {if(Length(img.Images) > 0) then begin
      img.nImages := 0; SetLength(img.Images, 0);
   end;}
end;

procedure imgDispose(var img: imgPImage);
begin
   if(img <> nil) then begin
      imgDispose(img^); dispose(img); img := nil;
   end;
end;

function imgValid(var img: imgTImage): boolean;
begin
   result := false;

   if(img.Image <> nil) then begin
      if(img.Width <> 0) and (img.Height <> 0) then result := true;
   end;
end;

{create a blank texture}
procedure imgCreateBlank(var img: imgTImage; w, h: longint; pf: longword);
begin
   img.Width := w;
   img.Height := h;
   img.PixF := pf;

   imgCalculate(img);
   imgAllocate(img);
end;

function imgMakeBlank(w, h: longint; pf: longword): imgPImage;
var
   p: imgPImage = nil;

begin
   p := imgMake();
   if(imgError = 0) then
      imgCreateBlank(p^, w, h, pf);

   result := p;
end;

{copy image from one to another}
procedure imgCopy(var src: imgTImage; var dst: imgTImage);
begin
   {dispose the previous destination image, if any}
   imgDispose(dst);

   {copy strings}
   dst.FileName := src.FileName;
   dst.FileName := src.sName;

   {copy data}
   dst := src;

   {NOTE: We wan't a pointer to a new image, not use the one from src}
   dst.Image := nil;

   {allocate memory for the new image data, if the source has an image}
   if(src.Image <> nil) then begin
      GetMem(dst.Image, dst.Size);
      if(dst.Image <> nil) then
         {copy the image data}
         move(src.Image^, dst.Image^, dst.Size)
      else begin imgeRaise(eNO_MEMORY); exit; end;
   end;

   {copy the palette too, if any and if not external}
   if(src.pal <> nil) and
      (src.Properties and imgcPROPERTIES_PAL_EXTERNAL = 0) then
         palCopy(src.pal^, dst.pal);
end;

procedure imgCopy(var src: imgTImage; var dst: imgPImage);
begin
   {dispose the previous destination image, if any}
   imgDispose(dst);

   {create a new image on the heap}
   dst := imgMake();
   if(imgError <> 0) then exit;

   {do the rest via the other routine}
   imgCopy(src, dst^);
end;

procedure imgCopyArea(var src, dst: imgTImage; px, py, dx, dy, w, h: longint);
var
   psize, len, i, algn, srcrowlen, dstrowlen: longint;
   p, s, d: pbyte;

begin
   if(imgValid(src) and imgValid(dst)) then begin
      if(w > 0) and (h > 0) then begin
         if(src.PixF = dst.PixF) then begin
            {get pixel depth}
            psize := imgPIXFDepth(src.PixF);
            if(psize mod 8 = 0) then begin
               psize := psize div 8; {get pixel size in bytes}
               len := psize * w; {figure out how many bytes to copy at a time}

               {check that the position does not go out of the boundaries}
               if(px+w > src.Width) or (dx+w >dst.Width) then begin
                  imgeRaise(eINVALID_ARG); exit;
               end else if(py+h > src.Height) or (dx+h > dst.Width) then begin
                  imgeRaise(eINVALID_ARG); exit;
               end;

               {calculate row lengths in bytes}
               srcrowlen := src.Width*psize+src.RowAlignBytes;
               dstrowlen := dst.Width*psize+dst.RowAlignBytes;

               {position at source}
               algn := src.RowAlignBytes*py;
               p := pbyte((py*src.Width+px)*psize+algn);
               s := src.Image+ptruint(p);

               {position at destination}
               algn := dst.RowAlignBytes*dy;
               p := pbyte((dy*dst.Width+dx)*psize+algn);
               d := pbyte(dst.Image)+ptruint(p);

               {copy rows one at a time}
               for i := 0 to (h-1) do begin
                  move(s^, d^, len);

                  {move to next row}
                  inc(s, srcrowlen);
                  inc(d, dstrowlen);
               end;
            end else imgeRaise(eUNSUPPORTED);
         end else imgeRaise(eUNSUPPORTED);
      end;
   end else imgeRaise(eINVALID);
end;

{ IMAGE INFORMATION }

function imgPIXFDepth(pf: longword): longint; inline;
begin
   if(pf <= PIXF_MAXIMUM) then imgPIXFDepth := imgcPIXFDepth[pf]
   else pf := 0;
end;

function imgPIXFChannels(pf: longword): longint; inline;
begin
   if(pf <= PIXF_MAXIMUM) then
      result := imgcPIXFChannels[pf]
   else result := 0;
end;

function imgPIXFChannelType(pf: longword): longint; inline;
begin
   if(pf <= PIXF_MAXIMUM) then
      result := imgcPIXFChannelTypes[pf]
   else result := PIXF_CHANNELS_INVALID;
end;

function imgPIXFHasAlpha(pf: longword): boolean; inline;
begin
   if(pf <= PIXF_MAXIMUM) then result := imgcPIXFHasAlpha[pf]
   else result := false;
end;

function imgPIXFColorChannelOrder(pf: longword): longint; inline;
begin
   if(pf <= PIXF_MAXIMUM) then
         result := imgcPIXFColorChannelOrder[pf]
   else result := -1;
end;

function imgPIXFName(pf: longword): string;
begin
   if(pf <= PIXF_MAXIMUM) then
         result := imgcPIXFNames[pf]
   else result := 'PIXF_UNKNOWN';
end;

{ IMAGE PROCESSING }

procedure imgFlipV(var img: imgTImage);
var
   i, lineSize, pos, pos2: longint;
   line: pointer = nil;
   PD: byte;

begin
   if(img.Image = nil) or (img.Width = 0) or (img.Height < 2) then exit;

   {get the pixel depth}
   PD := imgPIXFDepth(img.PixF);
   if(PD mod 8 <> 0) then begin imgeRaise(eUnsupported); exit; end;

   PD := PD div 8;

   {allocate enough memory for the line}
   linesize := (img.Width*PD);
   GetMem(line, linesize);
   if(line <> nil) then begin
      {flip all lines verticaly, that is swap them}
      pos := 0; pos2 := (img.Size-lineSize);
      for i := 0 to (img.Height div 2 - 1) do begin
         {put the first line into the temporary one}
         move((img.Image+pos)^, line^, lineSize);

         {put the second line into the first}
         move((img.Image+pos2)^, (img.Image+pos)^, lineSize);

         {put the temporary line into the second}
         move(line^, (img.Image+pos2)^, lineSize);

         inc(pos, lineSize); dec(pos2, lineSize);
      end;

      XFreeMem(line);
   end else imgeRaise(eNO_MEMORY);
end;

{TODO: Implementation of the routine is not completed.}
procedure imgFlipH(var img: imgTImage);
var
   i, lineSize, cPos: longword;
   line: pointer = nil;
   PD: byte;

begin
   if(img.Image = nil) or (img.Width = 0) or (img.Height = 0) then exit;

   {get the pixel depth}
   PD := imgPIXFDepth(img.PixF);
   if(PD mod 8 <> 0) then begin imgeRaise(eUnsupported); exit; end;

   PD := PD div 8;

   {allocate enough memory for the line}
   linesize := img.Width*PD;
   GetMem(line, linesize);
   if(line <> nil) then begin
      cPos := 0;

      {flip all lines horizontaly}
      for i := 0 to (img.Height-1) do begin
         {move the current line to temporary line memory}
         move((img.Image+cPos)^, line^, linesize);
         {now move the line back to it's place only flipped horizontaly}
      end;

      XFreeMem(line);
   end else imgeRaise(eNO_MEMORY);
end;

procedure imgTransform(var img: imgTImage; TargetFormat: byte);
var
   j, i: longint;

   newImage: pointer = nil;
   imgSize: longword;
   Depth, newDepth: byte;

   oiData, niData: pointer;
   oInc, nInc: longint;

   {NOTE: Colors stored in clr4ub and clr3ub must always have RGB/A channels,
   so the transformation routine needs not determine what channels are stored.}
   clr4ub: TColor4ub;
   clr3ub: TColor3ub absolute clr4ub;

{routines that convert from X to RGB/A accept a color and store it into clr4ub,
other routines use the color in clr4ub and store it into clr(parameter)}
procedure BGRtoRGB(var clr: TColor3ub);
begin
   {r=b, g=g, b=r}
   clr4ub[0] := clr[2]; clr4ub[1] := clr[1]; clr4ub[2] := clr[0];
end;

procedure RGBtoBGR(var clr: TColor3ub);
begin
   {r=b, g=g, b=r}
   clr[2] := clr4ub[0]; clr[1] := clr4ub[1]; clr[0] := clr4ub[2];
end;

procedure BGRAtoRGBA(var clr: TColor4ub);
begin
   clr4ub[0] := clr[2]; clr4ub[2] := clr[0]; clr4ub[1] := clr[1]; clr4ub[3] := clr[3];
end;

procedure RGBAtoBGRA(var clr: TColor4ub);
begin
   clr[2] := clr4ub[0]; clr[0] := clr4ub[2]; clr[1] := clr4ub[1]; clr[3] := clr4ub[3];
end;

procedure ARGBtoRGBA(var clr: TColor4ub);
begin
   {r=a, g=r, b=g, a=b}
   clr4ub[3] := clr[0]; clr4ub[0] := clr[1]; clr4ub[1] := clr[2]; clr4ub[2] := clr[3];
end;

procedure RGBAtoARGB(var clr: TColor4ub);
begin
   {r=a, g=r, b=g, a=b}
   clr[0] := clr4ub[3]; clr[1] := clr4ub[0]; clr[2] := clr4ub[1]; clr[3] := clr4ub[2];
end;

procedure ABGRtoRGBA(var clr: TColor4ub);
begin
   {r=a, g=b, b=g, a=r}
   clr4ub[3] := clr[0]; clr4ub[0] := clr[3]; clr4ub[1] := clr[2]; clr4ub[2] := clr[1];
end;

procedure RGBAtoABGR(var clr: TColor4ub);
begin
   clr[0] := clr4ub[3]; clr[3] := clr4ub[0]; clr[2] := clr4ub[1]; clr[1] := clr4ub[2];
end;

procedure cleanup();
begin
   {dispose of the old image}
   XFreeMem(img.Image);
   {dispose of the palette(if any)}
   palDispose(img);
end;

begin
   imgErrorReset();

   {first check the arguments}
   if(img.Image = nil) or (img.PixF = TargetFormat) then exit;

   if(img.Compression <> 0) or (img.PixelDepth = 0) then begin
      imgeRaise(eUnsupported); exit;
   end;

   {get information}
   Depth := imgPIXFDepth(img.PixF);
   newDepth := imgPIXFDepth(TargetFormat);

   {check if the depth is suitable}
   if(Depth mod 8 <> 0) or (newDepth mod 8 <> 0) then begin
      imgeRaise(eUNSUPPORTED); exit;
   end;

   {check, if the pixel format requires it, does the image have a palette}
   if(img.PixF = PIXF_INDEX_RGB_8) or
      (img.PixF = PIXF_INDEX_RGBA_8) then begin
      if(img.pal = nil) then begin imgeRaise(eINVALID); exit; end;
   end;

   {allocate enough space for the new image data}
   imgSize := (img.Height*img.Width)*imgPIXFDepth(img.PixF); {calculate size}
   GetMem(newImage, imgSize); {get memory}
   if(newImage <> nil) then begin 
      {get data}
      oiData := img.Image; {old image data}
      niData := newImage; {new image data}
      oInc := Depth div 8; {old image increment}
      nInc := newDepth div 8; {new image increment}

      for j := 0 to (img.Height-1) do begin
         for i := 0 to (img.Width-1) do begin
            {set the color to white with alpha 255}
            clr4ub := cWhite4ub;

            {get the old image pixel colors}
            case img.PixF of
               PIXF_INDEX_RGB_8:
                  clr3ub := TColorTable_8_RGB(img.pal^.data^)[byte(oiData^)];
               PIXF_INDEX_RGBA_8:
                  clr4ub := TColorTable_8_RGBA(img.pal^.data^)[byte(oiData^)];
               PIXF_RGB: clr3ub := TColor3ub(oiData^);
               PIXF_BGR: BGRtoRGB(TColor3ub(oiData^));
               PIXF_RGBA: clr4ub := TColor4ub(oiData^);
               PIXF_BGRA: BGRAtoRGBA(TColor4ub(oiData^));
               PIXF_ARGB: ARGBtoRGBA(TColor4ub(oiData^));
               PIXF_ABGR: ABGRtoRGBA(TColor4ub(oiData^));
            end;

            {transform to the new colors}
            case TargetFormat of
               PIXF_RGB: TColor3ub(niData^) := clr3ub;
               PIXF_BGR: RGBtoBGR(TColor3ub(niData^));
               PIXF_RGBA: TColor4ub(niData^) := clr4ub;
               PIXF_BGRA: RGBAtoBGRA(TColor4ub(niData^));
               PIXF_ARGB: RGBAtoARGB(TColor4ub(niData^));
               PIXF_ABGR: RGBAtoABGR(TColor4ub(niData^));
            end;

            {move to the next pixel}
            inc(oiData, oInc); inc(niData, nInc);
         end;
      end;

      cleanup();
      img.Image := newImage;

      {set the new pixel format}
      img.PixF := TargetFormat;
      imgCalculate(img);
   end else begin
      cleanup();
      imgeRaise(eNO_MEMORY);
   end;
end;

procedure imgGamma(var img: imgTImage; factor: single);
var
   PD, incr, i: longword;
   pos: pointer;
   scale, temp, r, g, b: single;

begin
   if(img.Image <> nil) then begin
      {check if the image is of a supported format}
      if(img.PixF <> PIXF_RGB) and (img.PixF <> PIXF_RGBA) and
        (img.PixF <> PIXF_BGR) and (img.PixF <> PIXF_BGRA) then exit;

      {get the pixel depth}
      PD := imgPIXFDepth(img.PixF);

      incr := PD div 8; pos := img.Image;
      for i := 0 to (img.Height*img.Width)-1 do begin
         scale := 1.0; temp := 0.0;

         {get the r,g,b values(or b,g,r it does not matter)}
         r := byte(pos^);
         g := byte((pos+1)^);
         b := byte((pos+2)^);

         {multiply the values by the factor, and keep the ratio of 255}
         r := r * factor / 255.0;
         g := g * factor / 255.0;
         b := b * factor / 255.0;

         if(r = 0) then temp := 0.0 else temp := 1.0 / r;
         if(r > 1.0) and (temp < scale) then scale := temp;
         if(g = 0) then temp := 0.0 else temp := 1.0 / g;
         if(g > 1.0) and (temp < scale) then scale := temp;
         if(b = 0) then temp := 0.0 else temp := 1.0 / b;
         if(b > 1.0) and (temp < scale) then scale := temp;

         {check if the values are higher than 255}

         {multiply all values by scale}
         scale := scale * 255.0;
         r := r * scale; g := g * scale; b := b * scale;

         {assign the new gamma corrected values}
         byte(pos^) := byte(round(r));
         byte((pos+1)^) := byte(round(g));
         byte((pos+2)^) := byte(round(b));

         pos := pos+incr;
      end;
   end;
end;

function imgSwapColorChannels(var img: imgTImage; chans: longword): boolean;
var
   i, incr, xincr: longint;
   cPixel: PColor4ub;

begin
   result := false;
   if(img.Image <> nil) then begin
      xincr := 0;
      {check if the image is of a supported pixel format}
      case img.PixF of
         PIXF_INDEX_RGB_4, PIXF_INDEX_RGB_8,
         PIXF_INDEX_RGBA_4, PIXF_INDEX_RGBA_8: begin
            palSwapColorChannels(img.pal^, chans); exit(true);
         end;
         PIXF_RGB, PIXF_BGR: incr := 3;
         PIXF_RGBA, PIXF_BGRA: incr := 4;
         PIXF_ARGB, PIXF_ABGR: begin incr := 4; xincr := 1; end;
         else exit; {unsupported format}
      end;

      {NOTE: xincr serves to skip channels that come before RGB or BGR channels}

      cPixel := pointer(img.Image);
      {Swap channels by using triple xor swapping. Faster than using normal
      swapping with a temp variable.}
      case chans of
         {swap red and green channels}
         imgccRG: begin
            for i := 0 to (img.Pixels-1) do begin
               inc(pointer(cPixel), xincr);
               cPixel^[1] := cPixel^[1] xor cPixel^[0];
               cPixel^[0] := cPixel^[1] xor cPixel^[0];
               cPixel^[1] := cPixel^[1] xor cPixel^[0];
               inc(pointer(cPixel), incr-xincr);
            end;
         end;
         {swap red and blue channels}
         imgccRB: begin
            for i := 0 to (img.Pixels-1) do begin
               inc(pointer(cPixel), xincr);
               cPixel^[2] := cPixel^[2] xor cPixel^[0];
               cPixel^[0] := cPixel^[2] xor cPixel^[0];
               cPixel^[2] := cPixel^[2] xor cPixel^[0];
               inc(pointer(cPixel), incr-xincr);
            end;
         end;
         {swap green and blue channels}
         imgccGB: begin
            for i := 0 to (img.Pixels-1) do begin
               inc(pointer(cPixel), xincr);
               cPixel^[1] := cPixel^[1] xor cPixel^[2];
               cPixel^[2] := cPixel^[1] xor cPixel^[2];
               cPixel^[1] := cPixel^[1] xor cPixel^[2];
               inc(pointer(cPixel), incr-xincr);
            end;
         end;
      end;
      result := true;
   end;
end;

procedure imgInvertColors(var img: imgTImage);
var
   i, Incr: longint;
   cPixel: PColor4ub;

begin
   if(img.Image <> nil) then begin
      {check if the image is of a supported pixel format}
      case img.PixF of
         PIXF_RGB, PIXF_BGR: incr := 3;
         PIXF_RGBA, PIXF_BGRA: incr := 4;
         else exit; {unsupported format}
      end;

      cPixel := pointer(img.Image);
      {invert the colors in the image}
      for i := 0 to (img.Pixels-1) do begin
         cPixel^[0] := 255 - cPixel^[0];
         cPixel^[1] := 255 - cPixel^[1];
         cPixel^[2] := 255 - cPixel^[2];
         inc(pointer(cPixel), Incr);
      end;
   end;
end;

procedure imgSetOrigin(var img: imgTImage; target: longint);
var
   target_vertical, target_horizontal: boolean;
   image_vertical, image_horizontal: boolean;

begin
   if(img.Origin <> target) then begin
      {get the origin settings}
      target_vertical := target and imgcBT_ORIGIN_VERTICAL > 0;
      target_horizontal := target and imgcBT_ORIGIN_HORIZONTAL > 0;
      image_vertical := img.Origin and imgcBT_ORIGIN_VERTICAL > 0;
      image_horizontal := img.Origin and imgcBT_ORIGIN_HORIZONTAL > 0;
      {flip the image if origins are different}
      if(target_vertical <> image_vertical) then imgFlipV(img);
      if(target_horizontal <> image_horizontal) then imgFlipH(img);

      img.Origin := target;
   end;
end;

{sets the image to the default origin, based on what it's origin is now}
procedure imgSetDefaultOrigin(var img: imgTImage);
begin
   imgSetOrigin(img, imgcORIGIN_DEFAULT);
end;

{set default colors}
procedure imgSetDefaultPixelFormat(var img: imgTImage);
var
   hasAlpha: boolean;

begin
   hasAlpha := imgPIXFHasAlpha(img.PixF);
   if(not hasAlpha) then
      imgTransform(img, imgcDefaultPixelFormat)
   else
      imgTransform(img, imgcDefaultPixelFormatAlpha);
end;

{set the image to the default color channel order}
procedure imgSetDefaultColorChannelOrder(var img: imgTImage);
var
   cco: longint;

begin
   {get the color channel order of iamge and exit if same as default}
   cco := imgPIXFColorChannelOrder(img.PixF);
   if(cco = PIXF_COLOR_CHANNEL_ORDER_DEFAULT) then exit;

   if(imgSwapColorChannels(img, imgccRB)) then
      case img.PixF of
         {RGB}
         PIXF_INDEX_RGB_4: img.PixF := PIXF_INDEX_BGR_4;
         PIXF_INDEX_RGB_8: img.PixF := PIXF_INDEX_BGR_8;
         PIXF_RGB: img.PixF := PIXF_BGR;
         PIXF_RGBA: img.PixF := PIXF_BGRA;
         PIXF_ARGB: img.PixF := PIXF_ABGR;
         {BGR}
         PIXF_INDEX_BGR_4: img.PixF := PIXF_INDEX_RGB_4;
         PIXF_INDEX_BGR_8: img.PixF := PIXF_INDEX_RGB_8;
         PIXF_BGR: img.PixF := PIXF_RGB;
         PIXF_BGRA: img.PixF := PIXF_RGBA;
         PIXF_ABGR: img.PixF := PIXF_ARGB;
      end;
end;

{ PALETTE }

procedure palInitRecord(var pal: imgTPalette); inline;
begin
   Zero(pal, SizeOf(imgTPalette));
end;

procedure palInitHandlerRecord(var palh: imgTPaletteHandler); inline;
begin
   Zero(palh, SizeOf(imgTPaletteHandler));
end;

procedure palInitLoaderDataRecord(var ld: palTLoaderData); inline;
begin
   Zero(ld, SizeOf(palTLoaderData));
end;

procedure palLoad(var pal: imgTPalette; const filename: string);
var
   data: palTLoaderData;
   fd: fhTFindData;
   fhdata: fhTData;
   f: TFile;
   fn: string;

   procedure logLoad();
   begin
      {$IFNDEF NO_LOGGING}
      if(not doLogNameAlways) then
         logWriteln(imgcName+' > (error)Image: '+FileName);
      {$ENDIF}
   end;

begin
   fInit(f);

   {initialize}
   imgErrorReset();
   if(length(filename) <> 0) then begin
      {set up the loader data}
      palInitLoaderDataRecord(data);
      data.f := @f;
      data.pal := @pal;

      {store the filename if specified}
      if(imgcStoreFileNames = true) then pal.fn := filename;
      
      {$IFNDEF NO_LOGGING}
      if(doLogNameAlways) then logWriteln(imgcName+' > Loading: '+FileName);
      {$ENDIF}

      {call the file handler which will figure out what loader to call}
      fhFindHandler(FileName, palLoadersInfo, fd);
      if(fhError <> 0) then begin
         imgeRaise(imgeFILEHANDLER); exit;
      end;

      {if a loader has been found}
      if(fd.handler <> nil) then begin
         {setup the filename}
         fn := filename;
         if(fhcGuessExtension) and (fd.ext <> nil) then
            fn := FileName+fd.ext^.ext;
         fhdata.fn := fn;

         {open the file if instructed to do so}
         if(not fd.handler^.DoNotOpenFile) then begin
            fOpen(f, fn);
            if(fError <> 0) then imgeRaise(eIO);

            data.fileSize := fSize(f);
            if(data.fileSize = 0) then begin
               fClose(f); imgeRaise(eEMPTY); exit;
            end;
         end;

         {call the loader}
         fhLoad(fd.handler^, @data, fhdata);

         {if there is no error, perform post-loading tasks}
         if(imgError = 0) then begin
         end;

         {close the file if instructed to do so}
         if(not fd.handler^.DoNotOpenFile) then begin
            fClose(f); fErrorIgn();
         end;
      end else begin imgeRaise(imgeLOADER_NOT_FOUND); exit; end;
   end else begin
      imgeRaise(eINVALID_ARG); exit;
   end;
end;

procedure palWrite(var pal: imgTPalette; const filename: string);
begin
   if(filename <> '') and (pal.Data <> nil) then begin
   end;
end;

procedure palInitData(var pal: imgTPalette);
var
   size: longint;

begin
   size := palSize(pal);
   if(pal.Data <> nil) then Zero(pal.Data^, size);
end;

{initialize}
procedure palMakeData(var pal: imgTPalette);
var
   size: longword;

begin
   size := palSize(pal);

   if(size > 0) then begin
      GetMem(pal.Data, size);
      if(pal.Data <> nil) then begin 
         pal.Size := size;
         palInitData(pal); 
      end else imgeRaise(eNO_MEMORY);
   end;
end;

procedure palMake(var img: imgTImage);
begin
   img.pal := palMake();
end;

procedure palMake(var img: imgTImage; pf: longword; nColors: longint);
begin
   palMake(img);
   if(imgError <> 0) then exit;

   img.pal^.PixF := pf;
   img.pal^.nColors := nColors;

   palMakeData(img.pal^);
   if(imgError <> 0) then exit;
end;

function palMake(): imgPPalette;
var
   pal: imgPPalette = nil;

begin
   new(pal);
   if(pal <> nil) then palInitRecord(pal^)
   else imgeRaise(eNO_MEMORY);

   result := pal;
end;

procedure palDispose(var pal: imgTPalette);
begin
   XFreeMem(pal.data);
   pal.sAuthor := '';
   pal.sName := '';
   pal.sDescription := '';
   pal.sProgram := '';
end;

procedure palDispose(var pal: imgPPalette);
begin
   if(pal <> nil) then begin
      palDispose(pal^); dispose(pal); pal := nil;
   end;
end;

procedure palDispose(var img: imgTImage);
begin
   if(img.Properties and imgcPROPERTIES_PAL_EXTERNAL = 0) and
      (img.pal <> nil) then palDispose(img.pal);
end;

function palValid(var img: imgTImage): boolean;
begin
   if(img.pal <> nil) then
      if(img.pal^.Data <> nil) then
         if(img.pal^.nColors > 0) then result := true
         else result := false
      else result := false
   else result := false
end;

function palValid(var pal: imgTPalette): boolean;
begin
   if(pal.Data <> nil) then
      if(pal.nColors > 0) then result := true
      else result := false
   else result := false
end;

function palSize(var pal: imgTPalette): longint;
var
   channels: longword;

begin
   result := 0;

   {get the properties of a palette}
   case pal.PixF of
      PIXF_RGB, PIXF_BGR: channels := 3;
      PIXF_RGBA, PIXF_BGRA, PIXF_ARGB, PIXF_ABGR: channels := 4;
      else exit;
   end;

   {calculate and return the size}
   result := channels*pal.nColors;
end;

{swaps the specified color channels in the palette}
procedure palSwapColorChannels(var pal: imgTPalette; chans: longword);
var
   i, elements, channels, Incr: longint;
   cPixel: PColor4ub;

begin
   if(not palValid(pal)) then exit;

   elements := pal.nColors;
   channels := imgPIXFChannels(pal.PixF);
   if(elements = 0) then exit;

   {get the first color in the palette}
   cPixel := pal.Data;

   {initialize swapping}
   Incr := channels;

   {Swap channels by using triple xor swapping. Faster than using normal
   swapping with a temp variable.}
   case chans of
      {swap red and green channels}
      imgccRG: begin
         for i := 0 to (elements-1) do begin
            cPixel^[1] := cPixel^[1] xor cPixel^[0];
            cPixel^[0] := cPixel^[1] xor cPixel^[0];
            cPixel^[1] := cPixel^[1] xor cPixel^[0];
            inc(pointer(cPixel), Incr);
         end;
      end;
      {swap red and blue channels}
      imgccRB: begin
         for i := 0 to (elements-1) do begin
            cPixel^[2] := cPixel^[2] xor cPixel^[0];
            cPixel^[0] := cPixel^[2] xor cPixel^[0];
            cPixel^[2] := cPixel^[2] xor cPixel^[0];
            inc(pointer(cPixel), Incr);
         end;
      end;
      {swap green and blue channels}
      imgccGB: begin
         for i := 0 to (elements-1) do begin
            cPixel^[1] := cPixel^[1] xor cPixel^[2];
            cPixel^[2] := cPixel^[1] xor cPixel^[2];
            cPixel^[1] := cPixel^[1] xor cPixel^[2];
            inc(pointer(cPixel), Incr);
         end;
      end;
      else exit;
   end;
end;

procedure palCopy(var src: imgTPalette; var dst: imgTPalette);
begin
   palDispose(dst);

   {copy strings}
   dst.fn := src.fn;
   dst.sAuthor := src.sAuthor;
   dst.sName := src.sName;
   dst.sDescription := src.sDescription;
   dst.sProgram := src.sProgram;

   dst := src;
   dst.Data := nil;

   {copy data}
   if(src.data <> nil) then begin
      GetMem(dst.data, dst.Size);
      if(dst.data <> nil) then move(src.data^, dst.data^, dst.Size)
      else imgeRaise(eNO_MEMORY);
   end;
end;

procedure palCopy(var src: imgTPalette; var dst: imgPPalette);
begin
   palDispose(dst);

   dst := palMake();
   if(imgError <> 0) then exit;

   palCopy(src, dst^);
end;

{ HELPER ROUTINES }
procedure imgCalculate(var img: imgTImage);
begin
   img.PixelDepth := imgPIXFDepth(img.PixF);

   {calculate the image size}
   img.Size := img.Width * img.Height * img.PixelDepth div 8;
   if(img.Width*img.Height mod 2 <> 0)  and (img.PixelDepth = 8) then inc(img.Size);

   {calculate the number of pixels}
   img.Pixels := img.Width*img.Height;
end;

procedure imgAllocate(var img: imgTImage);
begin
   GetMem(img.Image, img.Size);
   if(img.Image = nil) then imgeRaise(eNO_MEMORY);
end;

{ LOADER HELPER ROUTINES }

procedure imglCalculate(var ld: imgTLoaderData);
begin
   imgCalculate(ld.img^);
end;

procedure imglAllocate(var ld: imgTLoaderData);
begin
   imgAllocate(ld.img^);
end;

procedure imglReadPalette(var ld: imgTLoaderData; padBytes: longint);
var
   i, elements, channels, pos: longint;
   buf: array[0..31] of byte;
   pal: imgPPalette;

begin
   if(palValid(ld.img^)) and (ld.fileType = imgcFILE_TYPE_NORMAL) then begin
      pal := ld.img^.pal;

      {if there are no padding bytes we can read in the table directly}
      if(padBytes = 0) then begin
         imglBlockRead(ld, ld.img^.pal^.Data^, palSize(ld.img^.pal^));
         if(imgError <> 0) then exit;
      {otherwise read bit by bit and skip the padding bytes}
      end else begin
         {get the number of elements and channels}
         elements := pal^.nColors;
         channels := imgPIXFChannels(pal^.PixF);

         pos := 0;
         for i := 0 to (elements-1) do begin
            {read the current color and padding bytes}
            imglBlockRead(ld, buf, fileint(channels)+fileint(padBytes));
            if(imgError <> 0) then exit;

            {move the color into the palette}
            move(buf, (ld.img^.pal^.Data+pos)^, channels);
            inc(pos, channels);
         end;
      end;
   end;
end;

procedure imglProcessOrigin(var ld: imgTLoaderData);
begin
   if(imgcSetToDefaultOrigin) then
      imgSetDefaultOrigin(ld.img^);
end;

procedure imglProcessFormat(var ld: imgTLoaderData);
begin
   if(imgcSetToDefaultPixelFormat) then
      imgSetDefaultPixelFormat(ld.img^);
end;

procedure imglProcessColorChannelOrder(var ld: imgTLoaderData);
begin
   if(imgcSetToDefaultColorChannelOrder) then
      imgSetDefaultColorChannelOrder(ld.img^);
end;


procedure imglSeek(var ld: imgTLoaderData; pos: fileint);
begin
   if(ld.fileType = imgcFILE_TYPE_NORMAL) then begin
      fSeek(TFile(ld.f^), pos);
      if(fError <> 0) then begin imgeRaise(eIO); exit; end;
      ld.fPos := pos;
   end;
end;

function imglBlockRead(var ld: imgTLoaderData; var buf; size: fileint): fileint;
var
   rd: fileint = 0;

begin
   result := 0;
   if(ld.fileType = imgcFILE_TYPE_NORMAL) then begin
      result := fRead(TFile(ld.f^), buf, size);
      if(fError <> 0) then begin imgeRaise(eIO); end;
      inc(ld.fPos, rd);
   end;
end;

{reads in a image block}
function imglBlockWrite(var ld: imgTLoaderData; var buf; size: fileint): fileint;
var
   wd: fileint = 0;

begin
   result := 0;
   if(ld.fileType = imgcFILE_TYPE_NORMAL) then begin
      result := fWrite(TFile(ld.f^), buf, size);
      if(fError <> 0) then begin imgeRaise(eIO); end;
      inc(ld.fPos, wd);
   end;
end;

{ LOGGING }

{$IFNDEF NO_LOGGING}
procedure imgLogHandlers();
var
   curext: fhPExtension = nil;

procedure log(const s: string; start: fhPExtension);
var
   exts: string = '';

begin
   curext := start;
   if(curext <> nil) then repeat
      if(curext^.Next <> nil) then
         exts := exts + curext^.Ext+' '
      else
         exts := exts + curext^.ext;

      curext := curext^.Next;
   until (curext = nil);

   logWriteln(s + exts);
end;

begin
   logEnterSection('dImage');
   if(imgLoadersInfo.nExtensions > 0) then
      log('Loaders: ', imgLoadersInfo.eStart);
   if(imgWritersInfo.nExtensions > 0) then
      log('Writers: ', imgWritersInfo.eStart);
   logExitSection();
end;
{$ENDIF}

END.
